// Copyright (c) 2009, Greg Slepak. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
// 
// * Redistributions of source code must retain the above copyright notice, this
//   list of conditions and the following disclaimer.
// * Redistributions in binary form must reproduce the above copyright notice,
//   this list of conditions and the following disclaimer in the documentation
//   and/or other materials provided with the distribution.
// * Neither the name of the Greg Slepak nor the names of its contributors may be
//   used to endorse or promote products derived from this software without
//   specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
// ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

/*
 *  Common.h
 *  Espionage
 *
 *  Created by Greg Slepak on 4/22/08.
 *  Copyright 2008 Tao Effect LLC. All rights reserved.
 *
 */

#ifndef COMMON_H
#define COMMON_H

// -- logging --

#ifdef DEBUG
#	define log_debug(msg, ...) NSLog(@"DEBUG: " msg, ## __VA_ARGS__)
#else
#	define log_debug(msg, ...) /* nothing */
#endif

#ifdef DEBUG_SPECIAL
#	define log_special(msg, ...) NSLog(@"DEBUG: " msg, ## __VA_ARGS__)
#else
#	define log_special(msg, ...) /* nothing */
#endif

#define log_info(msg, ...) NSLog(@"INFO: " msg, ## __VA_ARGS__)
#define log_warn(msg, ...) NSLog(@"WARN (%s:%d): "  msg, __func__, __LINE__, ## __VA_ARGS__)
#define log_err(msg, ...)  NSLog(@"ERROR (%s:%d): " msg, __func__, __LINE__, ## __VA_ARGS__)
#define log_errp(msg, ...) NSLog(@"ERROR (%s:%d): " msg ": %s", __func__, __LINE__, ## __VA_ARGS__, strerror(errno))

// -- conveniences --

#define NSARY(...)							[NSArray arrayWithObjects:__VA_ARGS__, nil]
#define NSDCT(...)							[NSDictionary dictionaryWithObjectsAndKeys:__VA_ARGS__, nil]
#define NSINT(_n)							[NSNumber numberWithInt:_n]
#define NSYES								[NSNumber numberWithBool:YES]
#define NSNO								[NSNumber numberWithBool:NO]
#define NSSTR_FMT(_fmt, ...)				[NSString stringWithFormat:@_fmt, ## __VA_ARGS__]
#define NSERR(_msg, ...)					[NSError errorWithDomain:@"EspionageErrDomain" code:-1 userInfo:NSDCT(NSSTR_FMT(_msg, ## __VA_ARGS__), NSLocalizedDescriptionKey)]
#define ASSIGN(old, new)					do { id _old = (id)(old); (old) = [(new) retain]; [_old release]; } while (0)
#define DESTROY(var)						do { [(var) release]; (var) = nil; } while (0)
#define OBSERVE_NOTIF(_who, _what, _where)	[[NSNotificationCenter defaultCenter] addObserver:_who selector:_where name:_what object:nil]

// atomic ASSIGN and DESTROY

#define ASSIGN_ATOMIC(old, new) do { \
	id _o=(id)old, _n=[(id)new retain]; \
	while ( !OSAtomicCompareAndSwapPtr((void*)_o, (void*)_n, (void**)&old) ) \
		_o=(id)old; \
	[_o release]; \
} while (0)

#define DESTROY_ATOMIC(var)	do { \
	id _v=(id)var; \
	if ( OSAtomicCompareAndSwapPtr((void*)_v, NULL, (void**)&var) ) \
		[_v release]; \
} while (0)


#define JOIN_XY(a, b)						a##b
#define ACC_RETURN_H(type, var)				- (type)var;
#define ACC_RETURN_M(type, var)				- (type)var { return var; }
#define ACC_SET_H(type, var, Var)			- (void)JOIN_XY(set, Var):(type)JOIN_XY(a, var);
#define ACC_SET_M(type, var, Var)			- (void)JOIN_XY(set, Var):(type)JOIN_XY(a, var) { ASSIGN(var, JOIN_XY(a, var)); }
#define ACC_SETP_M(type, var, Var)			- (void)JOIN_XY(set, Var):(type)JOIN_XY(a, var) { var = JOIN_XY(a, var); }
#define ACC_COMBO_H(type, var, Var)			ACC_RETURN_H(type, var) ACC_SET_H(type, var, Var)
#define ACC_COMBO_M(type, var, Var)			ACC_RETURN_M(type, var) ACC_SET_M(type, var, Var)
#define ACC_COMBOP_M(type, var, Var)		ACC_RETURN_M(type, var) ACC_SETP_M(type, var, Var)

// -- lists --

// close to speed of obj-c 2.0's fast enum.
#define ENUMERATE(type, var, enumeratorMethod) \
	for (type e = (id)enumeratorMethod, \
		*imp = (id)[(NSEnumerator*)e methodForSelector:@selector(nextObject)], \
		*var = e ? ((IMP)imp)(e, @selector(nextObject)) : nil; \
		var; \
		var = ((IMP)imp)(e, @selector(nextObject)) \
	)

#define ENUMERID(var, enumeratorMethod) \
	for (id e = (id)enumeratorMethod, \
		imp = (id)[(NSEnumerator*)e methodForSelector:@selector(nextObject)], \
		var = e ? ((IMP)imp)(e, @selector(nextObject)) : nil; \
		var; \
		var = ((IMP)imp)(e, @selector(nextObject)) \
	)

// -- exception hanlding --

#define CATCH_EXCEPTION @catch (NSException *e) { log_warn("*** Caught %@! *** %@", [e name], [e reason]); }
#define TRY(...) @try { __VA_ARGS__; } CATCH_EXCEPTION

// -- error handling --

#define DO_FAILABLE(_errVar, _func, args...) do { \
	if ( (_errVar = _func(args)) != 0 ) { \
		log_err(#_func " returned: %d\n", (int)_errVar); \
		goto fail_label; \
	} \
} while (0)

#define DO_FAILABLE_SUB(_errVar, _subst, _func, args...) do { \
	if ( (_errVar = _func(args)) != 0 ) { \
		_errVar = _subst; log_err(#_func " resulted in: %d\n", (int)_errVar); \
		goto fail_label; \
	} \
} while (0)

#define FAILABLE(_errVar, _func, args...) do { \
	if ( (_errVar = _func(args)) != 0 ) { \
		log_err(#_func " returned: %d\n", (int)_errVar); \
	} \
} while (0)

#define FAIL_IF(_cond, args...) do { \
	if ( (_cond) != 0 ) { log_err("failed because: " #_cond); args; goto fail_label; } \
} while (0)

#define FAIL_IFQ(_cond, args...) do { if ( (_cond) != 0 ) { args; goto fail_label; } } while (0)

// -- NSCondition for 10.4 --

#if !defined(NSCondition) && !defined(MAC_OS_X_VERSION_10_5)
@interface NSCondition : NSObject <NSLocking> {}
- (void)wait;
- (BOOL)waitUntilDate:(NSDate *)limit;
- (void)signal;
- (void)broadcast;
@end
#endif

#endif /* COMMON_H */
