// Copyright (c) 2009, Greg Slepak. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
// 
// * Redistributions of source code must retain the above copyright notice, this
//   list of conditions and the following disclaimer.
// * Redistributions in binary form must reproduce the above copyright notice,
//   this list of conditions and the following disclaimer in the documentation
//   and/or other materials provided with the distribution.
// * Neither the name of the Greg Slepak nor the names of its contributors may be
//   used to endorse or promote products derived from this software without
//   specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
// ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#import <mach/task.h>
#import "TESharedObject.h"
#import "Common.h"

@implementation TESharedObject
- (id)initWithObject:(id)anObject
{
	kern_return_t err;
	
	if ( (self = [super init]) ) {
		DO_FAILABLE(err, semaphore_create, mach_task_self(), &sRead, SYNC_POLICY_FIFO, 1);
		DO_FAILABLE(err, semaphore_create, mach_task_self(), &sAccess, SYNC_POLICY_FIFO, 1);
		obj = [anObject retain];
	}
	return self;
	
fail_label:
	[self release];
	return nil;
}

- (void)dealloc
{
	semaphore_destroy(mach_task_self(), sRead);
	semaphore_destroy(mach_task_self(), sAccess);
	[obj release];
	[super dealloc];
}

- (id)borrowForWriting
{
	semaphore_wait(sAccess);
	writing = YES;
	return obj;
}

- (id)borrowForReading
{
	semaphore_wait(sRead);
	if ( ++readerCount == 1 )
		semaphore_wait(sAccess);
	semaphore_signal(sRead);
	return obj;
}

- (void)returnObject
{
	if ( writing )
	{
		writing = NO;
		semaphore_signal(sAccess);
	}
	else
	{
		semaphore_wait(sRead);
		if ( --readerCount == 0 )
			semaphore_signal(sAccess);
		semaphore_signal(sRead);
	}
}

- (void)setObject:(id)newObj
{
	semaphore_wait(sAccess);
	ASSIGN(obj, newObj);
	semaphore_signal(sAccess);
}
@end

@implementation TESharedMap
+ (TESharedMap *)map
{
	static TESharedMap *instance = nil;
	if ( !instance ) instance = [TESharedMap new];
	return instance;
}

- (id)init
{
	if ( (self = [super init]) ) {
		sharedMap = [[TESharedObject alloc] initWithObject:[NSMutableDictionary dictionary]];
	}
	return self;
}

- (void)dealloc
{
	[sharedMap release];
	[super dealloc];
}

- (id)borrowObjectForKey:(NSString *)key forReadingOnly:(BOOL)readonly
{
	TESharedObject *sharedObj = [[sharedMap borrowForReading] objectForKey:key];
	id obj = readonly ? [sharedObj borrowForReading] : [sharedObj borrowForWriting];
	[sharedMap returnObject];
	return obj;
}

- (void)returnObjectForKey:(NSString *)key
{
	[(TESharedObject *)[[sharedMap borrowForReading] objectForKey:key] returnObject];	
	[sharedMap returnObject];
}

- (void)setObject:(id)obj forKey:(NSString *)key
{
	NSMutableDictionary *dict = [sharedMap borrowForWriting];
	TESharedObject *sharedObj = [dict objectForKey:key];
	if ( !sharedObj )
	{
		sharedObj = [[TESharedObject alloc] initWithObject:obj];
		[dict setObject:sharedObj forKey:key];
		[sharedObj release];
	}
	else
	{
		[sharedObj setObject:obj];
	}
	[sharedMap returnObject];
}

- (void)removeObjectForKey:(NSString *)key
{
	NSMutableDictionary *dict = [sharedMap borrowForWriting];
	[dict removeObjectForKey:key];
	[sharedMap returnObject];
}

@end

